﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Programmable {
  [Item("ParameterVectorMoveGenerator", "Calls the GetNeighbors method of the programmable definition to obtain the moves.")]
  [StorableClass]
  public class ParameterVectorMoveGenerator : InstrumentedOperator, IParameterVectorMoveOperator, IMultiMoveGenerator, IStochasticOperator {

    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }

    public IValueLookupParameter<IntValue> SampleSizeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["SampleSize"]; }
    }

    public ILookupParameter<ISingleObjectiveProblemDefinitionHost> ProblemDefinitionParameter {
      get { return (ILookupParameter<ISingleObjectiveProblemDefinitionHost>)Parameters["ProblemDefinition"]; }
    }

    public ILookupParameter<Configuration> ConfigurationParameter {
      get { return (ILookupParameter<Configuration>)Parameters["Configuration"]; }
    }

    [StorableConstructor]
    protected ParameterVectorMoveGenerator(bool deserializing) : base(deserializing) { }
    protected ParameterVectorMoveGenerator(ParameterVectorMoveGenerator original, Cloner cloner)
      : base(original, cloner) { }
    public ParameterVectorMoveGenerator() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
      Parameters.Add(new ValueLookupParameter<IntValue>("SampleSize", "The number of moves to sample."));
      Parameters.Add(new LookupParameter<ISingleObjectiveProblemDefinitionHost>("ProblemDefinition", "The host that holds the problem definition."));
      Parameters.Add(new LookupParameter<Configuration>("Configuration", "An item that holds the problem's configuration."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ParameterVectorMoveGenerator(this, cloner);
    }

    public override IOperation InstrumentedApply() {
      var random = RandomParameter.ActualValue;
      var host = ProblemDefinitionParameter.ActualValue;
      if (host.Instance == null) throw new InvalidOperationException("Problem definition instance is null.");
      var sampleSize = SampleSizeParameter.ActualValue.Value;
      var config = ConfigurationParameter.ActualValue;
      var vector = Helper.Extract(ExecutionContext.Scope, config);
      var nbhood = host.Instance.GetNeighbors(random, vector).Take(sampleSize).ToList();

      var moveScopes = new Scope[nbhood.Count];
      for (int i = 0; i < moveScopes.Length; i++) {
        moveScopes[i] = new Scope(i.ToString());
        foreach (var param in nbhood[i].BinaryNames)
          moveScopes[i].Variables.Add(new Variable(param, nbhood[i].BinaryVector(param)));

        foreach (var param in nbhood[i].IntegerNames)
          moveScopes[i].Variables.Add(new Variable(param, nbhood[i].IntegerVector(param)));

        foreach (var param in nbhood[i].RealNames)
          moveScopes[i].Variables.Add(new Variable(param, nbhood[i].RealVector(param)));

        foreach (var param in nbhood[i].PermutationNames)
          moveScopes[i].Variables.Add(new Variable(param, nbhood[i].Permutation(param)));
      }
      ExecutionContext.Scope.SubScopes.AddRange(moveScopes);

      return base.InstrumentedApply();
    }
  }
}
